/*  Copyright 1987 Bart Miller.  All rights reserved.
 *  Explicit permission required to redistributed this software.
 */

/*==========================================================================*/
/*                                                                          */
/* Library routines for swb daemon interface.  Routines meet(),             */
/*      putserver() are contained in here.                                  */
/*                                                                          */
/* date       version   who          	change                              */
/* ---------  -------   -------------   ----------------------------------- */
/* 15 May 86    2.0     Stukenborg 	Replicated and reliable version     */
/* 15 Nov 85    1.0     Bart Miller  	Initial version                     */
/*                                                                          */
/*==========================================================================*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>

#include "swbmsg.h"
#include "table.h"

#define	TIMEOUT   10	
#define NUMTRIES   8
#define	ERROR     -2
#define	BADCOUNT  -1
#define ENDOFFILE  0
#define GOOD       1
#define TRUE       1


/*
 *		*------------------------------*
 *		|                              |
 *		|         setswbhost           |
 *		|                              |
 *		*------------------------------*
 *
 *----This is included only for backward compatibility.  Does nothing------
 */

int
setswbhost (hostname)
    char *hostname;
{
    return (SERV_OK);
}



/*
 *		*------------------------------*
 *		|                              |
 *		|            meet              |
 *		|                              |
 *		*------------------------------*
 */

int
meet (me, you)
    char *me;
    char *you;
{
    long servsock;
    long mysock;
    long status;
    int result;
    u_short myport, herport;
    servname hername;
    struct in_addr herhost;
    long CreateBoundSocket();
    long ConnectToServ();
    long ConnectToPeer();

    while ( TRUE ) {

    	if ((servsock = ConnectToServ ()) < 0) {
		fputs ("can't find an swb server\n", stderr);
		return(-1);
	}

    	if ((mysock = CreateBoundSocket (&myport)) < 0) {
		fputs ("can't create new port\n", stderr);
		exit (-1);
    	}

	result = ServRequest (SERV_PEER_MEET, servsock, me, you, myport,
			&status,
			&herhost, &herport, hername);
	close (servsock);


	/* if something went wrong with the server, try again */
	if (result < 0){
		close (mysock);
		continue;
	}

	/* if something really bad happened - bail out */
	if (status < 0){
		close (mysock);
		return (-1);
	}

	mysock = ConnectToPeer (mysock, status, herhost, herport);
    	return (mysock);
    }
}


/*
 *		*------------------------------*
 *		|                              |
 *		|         delserver            |
 *		|                              |
 *		*------------------------------*
 */

int
delserver (me)
    char *me;
{
    long servsock;
    long status;
    u_short herport;
    servname hername;
    struct in_addr herhost;
    int result;
    long ConnectToServ();
    long ConnectToPeer();

    while (TRUE) {

	if ((servsock = ConnectToServ ()) < 0) {
		fputs ("can't find an swb server\n", stderr);
		return(-1);
	}

    	result = ServRequest (SERV_DELETE, servsock, me, "", 0,
			&status,
			&herhost, &herport, hername);
    	close (servsock);

    	/* if something went wrong, try again */
	if (result < 0)
		continue;
	
	/* if something terrible happened - bail out */
	if (status < 0) {
		return (-1);
	}

	/*	ADDED BY PAUL */
	if (status == 3)
		continue;
    	return (0);
    }
}


/*
 *		*------------------------------*
 *		|                              |
 *		|         putserver            |
 *		|                              |
 *		*------------------------------*
 */

int
putserver (me)
    char *me;
{
    long servsock;
    long mysock;
    long status;
    u_short myport, herport;
    servname hername;
    struct in_addr herhost;
    int result;
    long ConnectToServ();
    long ConnectToPeer();

    if ((mysock = CreateBoundSocket (&myport)) < 0) {
	fputs ("can't create new port\n", stderr);
	exit (-1);
    }

    while (TRUE) {

	if ((servsock = ConnectToServ ()) < 0) {
		fputs ("can't find an swb server\n", stderr);
		return(-1);
	}

    	result = ServRequest (SERV_PUTSERVER, servsock, me, "", myport,
			&status,
			&herhost, &herport, hername);
    	close (servsock);

    	/* if something went wrong, try again */
	if (result < 0)
		continue;
	
	/* if something terrible happened - bail out */
	if (status < 0) {
		return (-1);
	}

	/*	ADDED BY PAUL */
	if (status == 3)
		continue;
		
	/***line removed 3/25/91: bart***/
	/***mysock = ConnectToPeer (mysock, status, herhost, herport);***/
    	return (mysock);
    }
}


/*
 *		*------------------------------*
 *		|                              |
 *		|        ServRequest           |
 *		|                              |
 *		*------------------------------*
 */

ServRequest(func, servsock, me, her, myport, status, herhost, herport, hername)
    int func;
    long servsock;
    servname me;
    servname her;
    u_short myport;
    long *status;
    struct in_addr *herhost;
    u_short *herport;
    servname hername;
{
    struct servrequest request;
    struct servresponse response;
    struct timeval to;
    servname hostname;
    int rv, count;
    fd_set mask;

    gethostname (hostname, SERVNAMELEN); 

    /*-----Send request to switchboard-------------------------*/

    request.function = func;

    strcpy (request.myname, me);
    request.myport = myport;
    request.hostid = 0;

    strcpy (request.hername, her);

    if (send (servsock, &request, sizeof (struct servrequest), 0) < 0) {
	return (-1);
    }

    /*-----Now wait for the response---------------------------*/

    /* set up timeout */
    to.tv_sec = TIMEOUT; /* seconds */
    to.tv_usec = 0; 	 /* micro-secs */

    /* set up readmask */
    FD_SET (servsock, &mask);

    count = select(NOFILE, &mask, 0, 0, &to);

    if (count == 0)
	return (-1); /* timed out, try again */
    
    if ((rv = GetResponse(servsock, &response) <= 0)) {
	return (-1);
    }

    herhost->s_addr = response.herhost.s_addr;
    (*herport) = response.herport;
    strcpy (hername, response.hername);
    (*status) = response.status;

    return (GOOD);
}


/*
 *		*------------------------------*
 *		|                              |
 *		|      CreateBoundSocket       |
 *		|                              |
 *		*------------------------------*
 */

long
CreateBoundSocket (myport) 
    u_short *myport;
{
    struct sockaddr_in sa;
    long sock, len;

    sock = socket (AF_INET, SOCK_STREAM, 0);

    sa.sin_family = AF_INET;
    sa.sin_addr.s_addr = INADDR_ANY;
    sa.sin_port = 0;

    if (bind (sock, &sa, sizeof (sa))) {
	perror ("binding");
	return (-1);
    }

    len = sizeof (sa);
    if (getsockname (sock, &sa, &len)) {
	perror ("getting sock name");
	return (-1);
    }

    *myport = sa.sin_port;
    return (sock);
}


/*
 *		*------------------------------*
 *		|                              |
 *		|      ConnectToPeer           |
 *		|                              |
 *		*------------------------------*
 */

long
ConnectToPeer (mysock, status, herhost, herport)
    long mysock;
    long status;
    struct in_addr herhost;
    u_short herport;
{
    struct hostent *gethostbyname();
    struct sockaddr_in sa;
    long newsock, len, tries=0;

    if (status == SERV_ACT)
    {
	/* active side */

	sa.sin_family = AF_INET;
	sa.sin_port = herport;
	sa.sin_addr.s_addr = herhost.s_addr;
again:
	if (connect (mysock, &sa, sizeof (struct sockaddr_in))) {
	    if (++tries < NUMTRIES) {
		sleep (2);  /* wait a bit first */
		goto again;
	    }
	    perror ("connecting to peer");
	    return (-1);
	}
	return (mysock);

    } else if (status == SERV_PAS) {

	/* passive side */
	
	len = sizeof (struct sockaddr_in);
	listen (mysock, 5); 
	if ((newsock = accept (mysock, &sa, &len)) < 0) { 
		perror ("on peer accept"); 
		return (-1);
	}
	close (mysock);
	return (newsock);

    } else {
#ifdef DEBUG
	fprintf (stderr, "bad return from swb (%d)\n", status);
#endif
	return (-1);
    }
}


/*
 *		*------------------------------*
 *		|                              |
 *		|      GetResponse             |
 *		|                              |
 *		*------------------------------*
 */

int
GetResponse(sock, response)
    long sock;
    struct servresponse *response;
{
    int rv;

    rv = read (sock, response, sizeof (struct servresponse));

    if (rv == 0)
	return (ENDOFFILE);
    else if (rv < 0) {
	perror ("error on ready reply");
	return (ERROR);
	}
    else if (rv != sizeof (struct servresponse))
	return (BADCOUNT);

#ifdef DEBUG
    printf("\nResponse:\n");
    printf(" status:  %d\n", response->status);
    printf(" hername: %s\n", response->hername);
    printf(" herport: %d\n", response->herport);
#endif
    return (GOOD);
}
